# Typing
from pydantic import BaseModel
from typing import List, Optional, Literal, Any
import logging
# Requests
import requests
import base64
# Utils
from mindgard.test import RequestHandler
from mindgard.utils import check_expected_args


class LabelConfidence(BaseModel):
    label: Optional[str]
    score: Optional[float]


class ImageModelWrapper:
    def __init__(self, url: str, labels: List[str], allow_redirects:bool = True, api_key: Optional[str] = None) -> None:
        self.url = url
        self.api_key = api_key
        self.labels = labels
        self._allow_redirects = allow_redirects

        # TODO: shape validation
        # self.shape = [3, 1, 224, 224]????


    def to_handler(self) -> RequestHandler:
        def handler(payload: Any) -> Any:
            logging.debug(f"received request {payload}")
            image_bytes = base64.b64decode(payload["image"])
            response = self(image=image_bytes)
            return {
                "response": [t.model_dump() for t in response]
            }
        return handler

    def __call__(self, image: bytes) -> List[LabelConfidence]:
        headers = {
            "Accept": "application/json",
            "Content-Type": "image/jpeg",
        }
        if self.api_key:
            headers["Authorization"] = f"Bearer {self.api_key}"

        res = requests.post(
            self.url, 
            headers=headers,
            data=image,
            allow_redirects=self._allow_redirects,
        )

        model_response = [LabelConfidence(label=item.get("label"), score=item.get("score")) for item in res.json()]

        aligned_data = image_label_tensor_align(self.labels, model_response)

        return aligned_data
    
def image_label_tensor_align(labels: List[str], model_data: List[LabelConfidence]) -> List[LabelConfidence]:
    if len(labels) < len(model_data):
        raise ValueError("Model returned more classes than the specified labels!")

    aligned: List[LabelConfidence] = []
    # make a list in the true order
    for true_label in labels:
        aligned.append(LabelConfidence(label=true_label, score=0.0))
    
    for label_confidence in model_data:
        if label_confidence.label not in labels:
            raise ValueError(f"Model returned an unknown class! ({label_confidence.label}) Doesn't match provided labels!")
    
    # where we have predictions, replace the score with the actual score
    for _, label_confidence in enumerate(model_data):
        for item in aligned:
            if item.label == label_confidence.label:
                item.score = label_confidence.score
                break
    return aligned



def get_image_model_wrapper(
    preset: Optional[Literal["huggingface", "local"]],
    api_key: Optional[str],
    url: str,
    labels: List[str],
    allow_redirects: bool = True, # backward compatibility
) -> ImageModelWrapper:

    # Isn't a concept of presets for image model wrapper as we just bind to HF
    if preset == "huggingface" or preset == None:
        check_expected_args(locals(), ["api_key", "url"])
        return ImageModelWrapper(api_key=api_key, url=url, labels=labels, allow_redirects=allow_redirects)  
    elif preset == "local":
        check_expected_args(locals(), ["url"])
        return ImageModelWrapper(url=url, labels=labels, allow_redirects=allow_redirects)



from typing import Dict
from rich.table import Table
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn, TaskID
from rich.live import Live
from mindgard.test import AttackState, Test, TestError
class TestUI():
  def __init__(self, test: Test, console:Console = Console()):
    self._console = console
    self._test = test

  def run(self) -> None:
    try:
      self._run()
    except TestError:
      # TODO: move error display to UI functions
      #       this is an interim step to stop the exception appearing twice.
      #       nb. we're only hiding errors we know to come from Test lib, not
      #       errors that might have come from the UI itself.
      pass

  def _run(self) -> None:
    test = self._test

    with test.state_wait_for(lambda state: state.submitting):
      with Progress(
        "{task.description}",
        SpinnerColumn(finished_text="[green3] Submitted!"),
        auto_refresh=True,
        console=self._console,
      ) as submit_progress:
        submit_task_id = submit_progress.add_task("Submitting...", start=True)

        with test.state_wait_for(lambda state: state.started):
          submit_progress.update(submit_task_id, completed=100)


    ##################################
    # the test is in progress
    progress_table = Table.grid(expand=True)

    overall_task_progress = Progress()
    exceptions_progress = Progress("{task.description}")
    
    attack_progress = Progress(
      "{task.description}",
      SpinnerColumn(finished_text="done"),
      TextColumn("{task.fields[status]}"),
    )

    progress_table.add_row(overall_task_progress)
    progress_table.add_row(attack_progress)
    progress_table.add_row("")
    progress_table.add_row(exceptions_progress)

    def task_update(attack_progress:Progress, task: TaskID, attack: AttackState) -> None:
      if attack.state == "completed":
        if attack.errored:
          attack_progress.update(task, status="[red3]failed", completed=1)
        else:
          attack_progress.update(task, status="[chartreuse3]success", completed=1)
      elif attack.state == "running":
        attack_progress.update(task, status="[yellow]running", completed=0)
      else:
        attack_progress.update(task, status="[chartreuse1]queued", completed=0)

    with test.state_wait_for(lambda state: (len(state.attacks) > 0 or state.test_complete)) as state:
      attack_id_task_map: Dict[str, TaskID] = {}
      for attack in state.attacks:
        attack_id_task_map[attack.id] = attack_progress.add_task(f"Attack {attack.name}", total=1)
        task_update(attack_progress, attack_id_task_map[attack.id], attack)

      attacks_progress = overall_task_progress.add_task("Progress", total=len(state.attacks))
      completed_attacks = sum(1 for attack in state.attacks if attack.state == "completed")
      overall_task_progress.update(attacks_progress, completed=completed_attacks)

      model_exception_task_map: Dict[str, TaskID] = {}
      model_exception_counts: Dict[str, int] = {}
      for model_exception in state.model_exceptions:
        name = model_exception
        if model_exception_task_map.get(name) is None:
          model_exception_task_map[name] = exceptions_progress.add_task("")
          model_exception_counts[name] = 1
        else:
          model_exception_counts[name] += 1

      for name, task in model_exception_task_map.items():
        exceptions_progress.update(task, description=f"[dark_orange3][!!!] {name} x{model_exception_counts[name]}")
      

    with Live(progress_table, refresh_per_second=10, console=self._console):
      while True:
        with test.state_then_wait_if(lambda state: not state.test_complete) as state:
          for attack in state.attacks:
            task = attack_id_task_map.get(attack.id)
            if task is None:  # library could have changed the attack list
              attack_id_task_map[attack.id] = attack_progress.add_task(f"Attack {attack.name}", total=1)
              task = attack_id_task_map[attack.id]
            task_update(attack_progress, task, attack)
          
          completed_attacks = sum(1 for attack in state.attacks if attack.state == "completed")
          overall_task_progress.update(attacks_progress, completed=completed_attacks)

          model_exception_counts: Dict[str, int] = {}
          for model_exception in state.model_exceptions:
            name = model_exception
            if model_exception_task_map.get(name) is None:
              model_exception_task_map[name] = exceptions_progress.add_task("")
            if model_exception_counts.get(name) is None:
              model_exception_counts[name] = 1
            else:
              model_exception_counts[name] += 1
          for name, task in model_exception_task_map.items():
            exceptions_progress.update(task, description=f"[dark_orange3][!!!] {name} x{model_exception_counts[name]}")
          
          if state.test_complete:
            break


      

    ##################################
    # the test is complete
    final_state = test.get_state()
    table = Table(title=f"Results - https://sandbox.mindgard.ai/r/test/{final_state.test_id}", width=80)
    table.add_column("Pass", style="cyan")
    table.add_column("Name", style="magenta")
    table.add_column("Risk", justify="right", style="green")

    for attack in final_state.attacks:
      if attack.errored:
        name = f"Error running '{attack.name}'"
        risk_str = "n/a"
        emoji = "❗️"
      else:
        name = attack.name
        risk_str = str(attack.risk)
        emoji = "✅️" if attack.passed else "❌‍"

      table.add_row(emoji, name, risk_str)

    self._console.print(table)


from typing import Any, Callable, Dict, Literal, Optional, Tuple, get_args
from rich.progress import TaskID, Progress
from rich.table import Table
from dataclasses import dataclass
import requests


@dataclass
class ExceptionCountTuple:
    task_id: TaskID
    count: int


# Web pub sub message types
type_wps_message_type = Literal["StartTest", "StartedTest", "Response", "Request"]

# UI task map stores a name for a task along with its taskid to be updated in progress object
type_ui_task_map = Dict[str, TaskID]
# Stores an exception message with the task to update, and the number of times the exception has been thrown
type_ui_exception_map = Dict[str, ExceptionCountTuple]

# interfaces that submit, polling and output functions MUST MEET to be usable in run_poll_display.
type_submit_func = Callable[[str, type_ui_exception_map, Progress], Any]
type_polling_func = Callable[
    [str, Any, type_ui_task_map, Progress],
    Optional[Any],
]
type_output_func = Callable[[Any, bool], Optional[Table]]

# Orchestrator submit/request enums
type_orchestrator_attack_pack = Literal["sandbox", "threat_intel"]
type_orchestrator_source = Literal["threat_intel", "user", "mindgard"]
model_types = ["image", "llm"]
type_model_types = Literal["image", "llm"]
type_model_presets = Literal['huggingface-openai', 'openai-compatible', 'huggingface', 'openai', 'azure-openai', 'azure-aistudio', 'anthropic', 'tester', 'local']
type_model_presets_list: Tuple[type_model_presets, ...] = get_args(type_model_presets)

# Types for dependency injecting get/post request functions into over api_post and api_get in orchestrator
type_post_request_function = Callable[[str, str, Dict[str, Any]], requests.Response]
type_get_request_function = Callable[[str, str], requests.Response]

# Different log levels
log_levels = [
    "critical",
    "fatal",
    "error",
    "warn",
    "warning",
    "info",
    "debug",
    "notset",
]  # [n.lower() for n in logging.getLevelNamesMapping().keys()]

# valid datasets
valid_image_datasets = [
    "mri",
    "xray",
    "rvltest_mini",
    "eurosat",
    "mnist",
    "beans",
]

valid_llm_datasets = {
    "customerservice": "BadCustomer",
    "finance": "BadFinance",
    "legal": "BadLegal",
    "medical": "BadMedical",
    "injection": "SqlInjection",
    "rce": "Xss",
    "xss": "Xss",
}


from ..types import (
    type_submit_func,
    type_ui_exception_map,
    type_ui_task_map,
)

from ..ui_prefabs import poll_and_display_test

from typing import Optional

from ..orchestrator import submit_sandbox_test, OrchestratorTestResponse

from rich.progress import Progress

POLL_INTERVAL_SECONDS = 3


def submit_sandbox_submit_factory(model_name: str) -> type_submit_func:

    def submit_sandbox_submit(
        access_token: str,
        ui_exception_map: type_ui_exception_map,
        ui_exception_progress: Progress,
    ) -> OrchestratorTestResponse:
        return submit_sandbox_test(access_token=access_token, target_name=model_name)

    return submit_sandbox_submit


def submit_sandbox_polling(
    access_token: str,
    initial_test: OrchestratorTestResponse,
    ui_task_map: type_ui_task_map,
    ui_task_progress: Progress,
) -> Optional[OrchestratorTestResponse]:
    return poll_and_display_test(
        access_token, ui_task_map, ui_task_progress, initial_test
    )



# Typing
import json
from typing import Optional
from .run_poll_display import type_ui_task_map

# Orchestrator
from .orchestrator import OrchestratorTestResponse, get_test_by_id, GetTestAttacksResponse

# Constants
from .constants import DASHBOARD_URL

# UI
from rich.progress import Progress
from rich.table import Table


def poll_and_display_test(
    access_token: str,
    ui_task_map: type_ui_task_map,
    ui_task_progress: Progress,
    initial_test: GetTestAttacksResponse,
) -> Optional[GetTestAttacksResponse]:
    test_and_attacks = get_test_by_id(access_token=access_token, test_id=initial_test.test.id)

    if len(ui_task_map.keys()) == 0:
        for item in test_and_attacks.items:
            ui_task_map[item.attack.id] = ui_task_progress.add_task(
                f"Attack {item.attack.attack_name}", total=1, status="[chartreuse1]queued"
            )

    for item in test_and_attacks.items:
        task_id = ui_task_map[item.attack.id]
        if item.attack.status == 2:
            ui_task_progress.update(task_id, completed=1, status="[chartreuse3]success")
        elif item.attack.status == -1:
            ui_task_progress.update(task_id, completed=1, status="[red3]failed")
        elif item.attack.status == 1:
            ui_task_progress.update(task_id, status="[orange3]running")

    if test_and_attacks.test.has_finished is False:
        return None
    return test_and_attacks


def output_test_table(
    json_out: bool,
    test: GetTestAttacksResponse,
    risk_threshold: int,
) -> Optional[Table]:
    if json_out:
        print(json.dumps(test.raw, indent=4))
        return None
    else:
        table = Table(title=f"Results - {DASHBOARD_URL}/r/test/{test.test.id}", width=80)
        table.add_column("Pass", style="cyan")
        table.add_column("Name", style="magenta")
        table.add_column("Flagged Events", justify="right", style="green")

        for item in test.items:
            if item.attack.status != 2:
                name = f"Error running '{item.attack.attack_name}'"
                risk_str = "n/a"
                emoji = "❗️"
            else:
                try:
                    flagged_to_total_ratio = item.attack.flagged_events / item.attack.total_events
                except ZeroDivisionError:
                    flagged_to_total_ratio = 0

                name = item.attack.attack_name
                risk_str = f"{item.attack.flagged_events} / {item.attack.total_events}"
                emoji = "❌‍" if flagged_to_total_ratio >= (risk_threshold / 100.0) else "✅️"

            table.add_row(emoji, name, risk_str)

        return table



"""
Temporary module to split off the __main__ for test execution using
new library system.
"""

from threading import Thread
from typing import Any, Dict, Union

import requests
from mindgard.auth import clear_token, load_access_token
from mindgard.constants import API_BASE, EXIT_CODE_ERROR, EXIT_CODE_NOT_PASSED, EXIT_CODE_PASSED
from mindgard.test import ImageModelConfig, LLMModelConfig, Test, TestConfig, UnauthorizedError
from mindgard.test_ui import TestUI
from mindgard.utils import print_to_stderr
from mindgard.wrappers.image import ImageModelWrapper
from mindgard.wrappers.llm import LLMModelWrapper


def run_test(final_args:Dict[str, Any], model_wrapper: Union[LLMModelWrapper, ImageModelWrapper]):
  access_token = load_access_token()
  if not access_token:
      print_to_stderr("\033[1;37mRun `mindgard login`\033[0;0m to authenticate.")
      exit(EXIT_CODE_ERROR)

  if final_args["model_type"] == "llm":
    test_config = TestConfig(
        api_base=API_BASE,
        api_access_token=access_token,
        target=final_args["target"],
        attack_source="user",
        dataset_domain=final_args.get('domain', None),
        parallelism=int(final_args["parallelism"]),
        model=LLMModelConfig(
            wrapper=model_wrapper,
            system_prompt=final_args["system_prompt"],
        ),
        exclude=final_args.get("exclude", None),
        include=final_args.get("include", None)
    )
    if (attack_pack := final_args.get("attack_pack")):
        test_config.attack_pack = attack_pack
  elif final_args["model_type"] == "image":
    test_config = TestConfig(
        api_base=API_BASE,
        api_access_token=access_token,
        target=final_args["target"],
        attack_source="user",
        parallelism=int(final_args["parallelism"]),
        model=ImageModelConfig(
            wrapper=model_wrapper,
            dataset=final_args["dataset"],
            model_type=final_args["model_type"],
            labels=final_args["labels"],
        )
    )
  else:
    raise NotImplementedError("Invalid model type") # This should be unreachable due to argparse validations
  
  test = Test(test_config)
  test_ui = TestUI(test)

  # daemonize to prevent blocking the main thread's exit if test.run raises exception
  test_ui_thread = Thread(target=test_ui.run, name="TestUI")
  test_ui_thread.start()

  try:
      test.run()
  except UnauthorizedError:
      print_to_stderr(
          "Access token is invalid. Please re-authenticate using `mindgard login`"
      )
      clear_token()
      exit(EXIT_CODE_ERROR)
  except requests.HTTPError as e:
      if "Unauthorized" in str(e):
          print_to_stderr(
              "Access token is invalid. Please re-authenticate using `mindgard login`"
          )
          clear_token()
          exit(EXIT_CODE_ERROR)
      else:
          raise e

  test_ui_thread.join()
  
  if test.get_state().passed:
    exit(EXIT_CODE_PASSED)
  elif test.get_state().passed == False:
    exit(EXIT_CODE_NOT_PASSED)


from typing import Dict, Optional, List
import re

def parse_headers(headers_comma_separated: Optional[str] = None, headers_list: List[str] = []) -> Dict[str, str]:
    if headers_comma_separated is None:
        return parse_list_of_headers(headers_input=headers_list or [])
    else:
        return parse_list_of_headers(headers_input=headers_comma_separated.split(",") + (headers_list or []))

def parse_list_of_headers(headers_input:list[str]) -> Dict[str, str]:
    headers: Dict[str, str] = {}
    for key_and_value_str in headers_input:
        match = re.search("(.*?): *(.*)", key_and_value_str)
        if match:
            key = match.group(1)
            value = match.group(2)
            headers[key.strip()] = value.strip()

    return headers


# Types
from typing import Dict, Any

# Requests
import requests
from tenacity import retry, stop_after_attempt, wait_fixed

# Constants
from .constants import (
    VERSION,
    API_RETRY_ATTEMPTS,
    API_RETRY_WAIT_BETWEEN_ATTEMPTS_SECONDS,
)


def _standard_headers(access_token: str) -> Dict[str, str]:
    return {
        "Authorization": f"Bearer {access_token}",
        "User-Agent": f"mindgard-cli/{VERSION}",
        "X-User-Agent": f"mindgard-cli/{VERSION}",
    }


@retry(
    stop=stop_after_attempt(API_RETRY_ATTEMPTS),
    wait=wait_fixed(API_RETRY_WAIT_BETWEEN_ATTEMPTS_SECONDS),
    reraise=True,
)
def api_post(url: str, access_token: str, payload: Dict[str, Any]) -> requests.Response:
    response = requests.post(
        url=url, json=payload, headers=_standard_headers(access_token)
    )
    response.raise_for_status()
    return response


@retry(
    stop=stop_after_attempt(API_RETRY_ATTEMPTS),
    wait=wait_fixed(API_RETRY_WAIT_BETWEEN_ATTEMPTS_SECONDS),
    reraise=True,
)
def api_get(url: str, access_token: str) -> requests.Response:
    response = requests.get(url=url, headers=_standard_headers(access_token))
    response.raise_for_status()
    return response



"""
Status: Pre-Alpha -- implementation incomplete and interfaces are likely to change

Provides headless execution of mindgard tests.
"""

import contextlib
import copy
from dataclasses import dataclass, field
import logging
from threading import Condition
import time
from typing import Any, Callable, Dict, Literal, Optional, Protocol, Tuple, List
from azure.messaging.webpubsubclient import WebPubSubClient, WebPubSubClientCredential
from azure.messaging.webpubsubclient.models import OnGroupDataMessageArgs, CallbackType, WebPubSubDataType
import requests
from mindgard.constants import DEFAULT_RISK_THRESHOLD
from mindgard.exceptions import handle_exception_callback
from mindgard.mindgard_api import AttackResponse, MindgardApi, FetchTestAttacksData

from mindgard.version import VERSION

# surpress distracting warning messages from azure webpubsubclient; aims to be as precise as possible
# https://github.com/Azure/azure-sdk-for-python/issues/37390
logging.getLogger('azure.messaging.webpubsubclient._client').addFilter(
    lambda record: not(record.getMessage() == "The client is stopping state. Stop recovery." and record.levelno == logging.WARNING)
)

class TestError(Exception):
    pass

class InternalError(TestError):
    pass

class UnauthorizedError(TestError):
    pass

class RequestHandler(Protocol):
    """
    Protocol for converting a Request paylod to a Response payload

    Receives payload dict and returns reponse payload (to send unaltered) and an optional error string
    """
    def __call__(self, payload: Dict[str, Any]) -> Dict[str, Any]: ...

class _Wrapper(Protocol):
    """
    Protocol for supplied model wrappers
    """

    def to_handler(self) -> RequestHandler: ...

@dataclass
class ModelConfig:
    wrapper: _Wrapper
    def to_orchestrator_init_params(self) -> Dict[str, Any]:
        return {}
    
@dataclass
class ImageModelConfig(ModelConfig):
    dataset: str
    labels: list[str]
    model_type: str = "image"

    def to_orchestrator_init_params(self) -> Dict[str, Any]:
        return {
            "modelType": self.model_type,
            "dataset": self.dataset,
            "labels": self.labels,
        }

@dataclass
class LLMModelConfig(ModelConfig):
    system_prompt: str
    model_type: str = "llm"

    def to_orchestrator_init_params(self) -> Dict[str, Any]:
        return {
            "modelType": self.model_type,
            "system_prompt": self.system_prompt,
        }

@dataclass
class TestConfig:
    api_base: str
    api_access_token: str
    target: str
    attack_source: str
    parallelism: int
    model: ModelConfig
    target_id: Optional[str] = None
    dataset_domain: Optional[str] = None
    attack_pack: str = "sandbox"
    additional_headers: Optional[Dict[str, str]] = None
    exclude: Optional[List[str]] = None
    include: Optional[List[str]] = None
    risk_threshold: int = DEFAULT_RISK_THRESHOLD
    def to_orchestrator_init_params(self) -> Dict[str, Any]:
        """
        Get parameters for the init test request to orchestrator
        """
        params = {**{
            "target": self.target,
            "attackPack": self.attack_pack,
            "parallelism": self.parallelism,
            "attackSource": self.attack_source
        }, **self.model.to_orchestrator_init_params()}

        if self.dataset_domain is not None:
            params['datasetDomain'] = self.dataset_domain

        if self.target_id is not None:
            params['target_id'] = self.target_id
        
        if self.exclude is not None:
            params['exclude'] = self.exclude
        
        if self.include is not None:
            params['include'] = self.include

        return params

    def handler(self) -> RequestHandler:
        return self.model.wrapper.to_handler()

@dataclass
class AttackState():
    id: str
    name: str
    state: Literal["queued", "running", "completed"]
    errored: Optional[bool] = None
    passed: Optional[bool] = None
    risk: Optional[int] = None

@dataclass
class TestState():
    submitting: bool = False
    submitted: bool = False
    started: bool = False
    test_complete: bool = False
    attacks: List[AttackState] = field(default_factory=list[AttackState])
    model_exceptions: List[str] = field(default_factory=list[str])
    test_id: Optional[str] = None
    passed: Optional[bool] = None

    def clone(self):
        return TestState(
            submitting=self.submitting,
            submitted=self.submitted,
            started=self.started,
            test_complete=self.test_complete,
            attacks=copy.deepcopy(self.attacks),
            model_exceptions=[exception for exception in self.model_exceptions],
            test_id=self.test_id,
            passed=self.passed
        )
class TestImplementationProvider():

    def __init__(self, mindgard_api:Optional[MindgardApi] = None):
        self._mindgard_api = mindgard_api or MindgardApi()

    def init_test(self, config:TestConfig) -> Tuple[str, str]:
        """
        Init a test in with API and return the url and group_id
        """
        url = f"{config.api_base}/tests/cli_init"

        response = requests.post(
            url=url, 
            headers={
                "Authorization": f"Bearer {config.api_access_token}",
                "User-Agent": f"mindgard-cli/{VERSION}",
                "X-User-Agent": f"mindgard-cli/{VERSION}",
                **(config.additional_headers or {})
            },
            json=config.to_orchestrator_init_params()
        )
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if e.response.status_code == 401:
                raise UnauthorizedError("Invalid access token") from e
            elif e.response.status_code == 403:
                raise UnauthorizedError("Access denied") from e
            else:
                raise InternalError("An unexpected error occurred during the test initialization") from e
        except Exception as e:
            raise InternalError("An unexpected error occurred during the test initialization") from e

        url = response.json().get("url", None)
        group_id = response.json().get("groupId", None)
        if not url or not group_id:
            raise ValueError("Invalid response from orchestrator, missing websocket credentials.")

        return url, group_id
    
    def create_client(self, connection_url:str) -> WebPubSubClient:
        credentials = WebPubSubClientCredential(client_access_url_provider=connection_url)
        return WebPubSubClient(credential=credentials)

    def connect_websocket(self, client:WebPubSubClient) -> None:
        client.open()

    def register_handler(self, handler:RequestHandler, client:WebPubSubClient, group_id:str) -> None:
        def callback(msg:OnGroupDataMessageArgs) -> None:
            if msg.data["messageType"] != "Request":
                return

            payload = handler(payload=msg.data["payload"])
            
            client.send_to_group(
                "orchestrator",
                {
                    "correlationId": msg.data["correlationId"],
                    "messageType": "Response",
                    "status": "ok",
                    "payload": payload
                },
                data_type=WebPubSubDataType.JSON
            )
            
        client.subscribe(CallbackType.GROUP_MESSAGE, callback)

    def start_test(self, client:WebPubSubClient, group_id:str, timeout:float = 10) -> str:
        logging.info(f"start_test: opening connection and starting test")
        started_condition = Condition()
        test_ids: list[str] = []
        def handler(msg:OnGroupDataMessageArgs) -> None:
            logging.debug(f"received message {msg.data}")
            if msg.data["messageType"] == "StartedTest":
                with started_condition:
                    test_ids.append(msg.data["payload"]["testId"])
                    started_condition.notify_all()

        client.subscribe(CallbackType.GROUP_MESSAGE, handler)
        payload:Dict[str, Any] = {
            "correlationId": "",
            "messageType": "StartTest",
            "payload": {"groupId": group_id},
        }
        client.send_to_group(group_name="orchestrator", content=payload, data_type=WebPubSubDataType.JSON)
        
        logging.info("start_test: waiting for test_id")
        with started_condition:
            found = started_condition.wait_for(lambda: len(test_ids) > 0, timeout=timeout)
            if not found:
                raise InternalError("Failure: timeout waiting for test to start")
            test_id = test_ids[0]
            logging.info(f"start_test: received test_id {test_id}")
            return test_id

    def poll_test(self, config:TestConfig, test_id:str) -> Optional[FetchTestAttacksData]:
        try:
            return self._mindgard_api.fetch_test_attacks(
                api_base=config.api_base,
                access_token=config.api_access_token,
                additional_headers=config.additional_headers,
                test_id=test_id
            )
        except Exception as e:
            raise InternalError("An unexpected error occurred during the test polling") from e

    def close(self, client:Optional[WebPubSubClient]) -> None:
        if client:
            client.close()

def _attack_response_to_attack_state(attack_data:AttackResponse, risk_threshold: int) -> AttackState:
    return AttackState(
        id=attack_data.id,
        name=attack_data.name,
        state=attack_data.state,
        errored=attack_data.errored,
        passed=None if attack_data.risk is None else (attack_data.risk < risk_threshold),
        risk=None if attack_data.errored else attack_data.risk
    )

class Test():
    def __init__(self, config:TestConfig, poll_period_seconds:float = 5):
        self._config = config
        self._state = TestState() # TOOD: coverage
        self._provider = TestImplementationProvider() # TOOD: coverage
        self._notifier = Condition()
        self._poll_period_seconds = poll_period_seconds
        self._exit_error = None

    # for clients to observe state[changes]
    def get_state(self) -> TestState:
        with self._notifier:
            return self._state.clone()
        
    @contextlib.contextmanager
    def state_then_wait_if(self, predicate:Callable[[TestState], bool]):
        with self._notifier:
            self._raise_on_error()
            yield self.get_state()
            if predicate(self._state):
                self._notifier.wait()
                self._raise_on_error()

    # TODO: coverage
    @contextlib.contextmanager
    def state_wait(self):
        with self._notifier:
            self._raise_on_error()
            self._notifier.wait()
            self._raise_on_error()
            yield self._state

    # TODO: coverage
    @contextlib.contextmanager
    def state_wait_for(self, predicate:Callable[[TestState], bool]):
        with self._notifier:
            self._notifier.wait_for(lambda: self._exit_error or predicate(self._state))
            self._raise_on_error()
            yield self._state

    def _set_started(self) -> None:
        with self._notifier:
            self._state.started = True
            self._notifier.notify_all()

    def _set_submitting_test(self) -> None:
        with self._notifier:
            self._state.submitting = True
            self._notifier.notify_all()

    def _set_attacking(self, test_id:str, attacks:List[AttackState]) -> None:
        with self._notifier:
            self._state.submitted = True
            self._state.test_id = test_id
            self._state.attacks = attacks
            self._state.started = True
            self._notifier.notify_all()

    def _set_test_complete(self, test_id:str, attacks:List[AttackState], risk:int) -> None:
        with self._notifier:
            self._state.test_id = test_id
            self._state.attacks = attacks
            self._state.test_complete = True
            self._state.passed = risk < self._config.risk_threshold
            self._notifier.notify_all()

    def _add_exception(self, exception:Exception) -> None:
        with self._notifier:
            self._state.model_exceptions.append(str(exception))
            self._notifier.notify_all()

    def _set_error(self, exception:BaseException) -> None:
        with self._notifier:
            self._exit_error = exception
            self._notifier.notify_all()

    def _raise_on_error(self) -> None:
        """
        Raise an exception if an error has occurred
        """
        if self._exit_error:
            raise self._exit_error

    # run the test
    def run(self) -> None:
        self._set_started()
        p = self._provider
        wps_client = None
        test_id = None
        handler = self._config.handler()

        # TODO: I don't want to wrap the wrapped wrapper like this
        #       but was too nervous to change the internals of model wrappers
        #       until some of the old code is gone.
        def my_handler(payload:Dict[str, Any]) -> Dict[str,Any]:
            try:
                return handler(payload)
            except Exception as e:
                self._add_exception(e)
                # logging.error(f"Error in handler: {e}") # ??????
                
                # arguably we could pass self._add_exception to this, but the callback
                # was not tested, and suspect the direction of travel is against that
                error_code = handle_exception_callback(e, None)
                return {
                    "response": "",
                    "error": error_code,
                }

        try:
            self._set_submitting_test()
            wps_url, group_id = p.init_test(self._config)
            wps_client = p.create_client(wps_url)
            p.connect_websocket(wps_client)
            p.register_handler(my_handler, wps_client, group_id)
            test_id = p.start_test(wps_client, group_id)

            finished = False
            retries = 0
            while not finished:
                try:
                    retries += 1
                    test_data = p.poll_test(
                        config=self._config,
                        test_id=test_id
                    )
                except Exception as e:
                    if retries > 9:
                        raise
                else:
                    if test_data is not None:
                        retries = 0
                        finished = test_data.has_finished
                        if finished:
                            self._set_test_complete(test_id, [], 0)
                        else:
                            self._set_attacking(test_id, [])
                    elif retries > 9:
                        raise InternalError("Failed to poll test data 10 times")
                if not finished:
                    time.sleep(self._poll_period_seconds)
        except TestError as e:
            self._set_error(e)
            raise e
        except Exception as e:
            self._set_error(e)
            raise InternalError("An unexpected error occurred during the test execution") from e
        except BaseException as e:
            self._set_error(e)
            raise
        finally:
            p.close(wps_client)


from mindgard.config import load_auth_config
 
VERSION: str = "0.74.0"
REPOSITORY_URL: str = f"https://pypi.org/pypi/mindgard/json"

ALGORITHMS = ['RS256']

auth_config = load_auth_config()
AUTH0_DOMAIN = auth_config.AUTH0_DOMAIN
AUTH0_CLIENT_ID = auth_config.AUTH0_CLIENT_ID
AUTH0_AUDIENCE = auth_config.AUTH0_AUDIENCE
API_BASE = auth_config.API_BASE
DASHBOARD_URL = auth_config.DASHBOARD_URL

API_RETRY_ATTEMPTS = 10
API_RETRY_WAIT_BETWEEN_ATTEMPTS_SECONDS = 3
DEFAULT_RISK_THRESHOLD = 50

EXIT_CODE_PASSED = 0
EXIT_CODE_NOT_PASSED = 1
EXIT_CODE_ERROR = 2

ATTACK_STATE_QUEUED = 0
ATTACK_STATE_RUNNING = 1
ATTACK_STATE_COMPLETED = 2
ATTACK_STATE_ERRORED = -1


import os
from typing import Any, Callable, Optional, TypeVar, cast
from functools import wraps

import requests
from auth0.authentication.token_verifier import (
    AsymmetricSignatureVerifier,  # type: ignore
    TokenVerifier,
)

from .utils import CliResponse, print_to_stderr
from .config import (
    get_token_file,
    get_instance_file,
    load_auth_config,
)
from .config import (
    is_instance_set,
    instance_auth_config,
    sandbox_auth_config,
    create_config_directory,
)
import requests
from .utils import print_to_stderr

from rich.console import Console

import json
import time


def clear_token() -> None:
    if os.path.exists(get_token_file()):
        os.remove(get_token_file())


def clear_instance() -> None:
    """
    Deletes the instance file
    """
    if os.path.exists(get_instance_file()):
        os.remove(get_instance_file())


def login(instance: str) -> None:
    """
    Runs the device authorization flow and stores the user token in memory.
    If instance is set by user, grab required instance variables from their deployment.

    :param instance: string
    """

    if not is_instance_set():
        print("Welcome to Mindgard! Let's get you authenticated...")
        print(
            "\033[1;37mNote: Mindgard is an AI security testing tool that will run red-team attacks to assess the risk of the AI systems you are testing."
        )
        print(
            "Only use Mindgard with systems you have authorization to test in this manner.\033[0;0m\n"
        )
        print("By continuing you acknowledge this and the terms of service.\n")

        if instance:
            customer_instance_response = requests.get(
                "https://api.{}.mindgard.ai/api/v1/cli/context".format(instance)
            )

            if customer_instance_response.status_code != 200:
                print_to_stderr(
                    "Error communicating with deployed instance. Please validate instance name and try again. Contact Mindgard support if the issue persists."
                )
                raise ValueError(
                    f"Error communicating with deployed instance: {customer_instance_response.json()}"
                )

            customer_instance_data = customer_instance_response.json()

            auth_configs = instance_auth_config(customer_instance_data)
        else:
            auth_configs = sandbox_auth_config()

        device_code_payload = {
            "client_id": auth_configs.AUTH0_CLIENT_ID,
            "scope": "openid profile email offline_access",
            "audience": auth_configs.AUTH0_AUDIENCE,
        }
        device_code_response = requests.post(
            "https://{}/oauth/device/code".format(auth_configs.AUTH0_DOMAIN),
            data=device_code_payload,
        )

        if device_code_response.status_code != 200:
            print_to_stderr(
                "Error generating login url. Please try again. Contact Mindgard support if the issue persists."
            )
            raise ValueError(
                f"Login service API response: {device_code_response.json()}"
            )

        device_code_data = device_code_response.json()
        print(
            "1. On your computer or mobile device navigate to: ",
            device_code_data["verification_uri_complete"],
        )
        print(
            "2. Confirm that you see the following code: ",
            device_code_data["user_code"],
        )
        print("3. Register/log in using the web UI\n")

        # New code 👇
        token_payload = {
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": device_code_data["device_code"],
            "client_id": auth_configs.AUTH0_CLIENT_ID,
            "audience": auth_configs.AUTH0_AUDIENCE,
        }

        console = Console()
        with console.status("[bold green]Waiting for auth to complete...") as status:
            authenticated = False
            while not authenticated:
                token_response = requests.post(
                    "https://{}/oauth/token".format(auth_configs.AUTH0_DOMAIN),
                    data=token_payload,
                )

                token_data = token_response.json()
                if token_response.status_code == 200:
                    validate_id_token(
                        token_data["id_token"],
                        auth_configs.AUTH0_DOMAIN,
                        auth_configs.AUTH0_CLIENT_ID,
                    )
                    print("Authenticated!")
                    # Create config directory
                    create_config_directory()

                    with open(get_token_file(), "w") as f:
                        f.write(token_data["refresh_token"])

                    if instance:
                        # Write instance variables to config dir
                        with open(get_instance_file(), "w") as f:
                            json.dump(customer_instance_data, f)

                    authenticated = True
                elif token_data["error"] not in ("authorization_pending", "slow_down"):
                    error = token_data.get(
                        "error_description",
                        "Error authenticating the user. Please wait 30s and try again.",
                    )
                    raise ValueError(error)
                else:
                    time.sleep(device_code_data["interval"])
    else:
        raise ValueError("Please logout of current instance with `mindgard logout`")


def logout() -> None:
    """
    Removes the user token
    """
    print(f"Clearing credentials stored at {get_token_file()}.")
    clear_token()

    if is_instance_set():
        print(f"Clearing credentials stored at {get_instance_file()}.")
        clear_instance()

    print("Logged out!")


# TODO: test
def validate_id_token(id_token: str, auth0_domain: str, auth0_client_id: str) -> None:
    """
    Verify the token and its precedence. Raises if the token is invalid.

    :param id_token: string
    """
    jwks_url = "https://{}/.well-known/jwks.json".format(auth0_domain)
    issuer = "https://{}/".format(auth0_domain)
    sv = AsymmetricSignatureVerifier(jwks_url)
    tv = TokenVerifier(signature_verifier=sv, issuer=issuer, audience=auth0_client_id)
    tv.verify(id_token)


def load_access_token() -> Optional[str]:
    """
    Reads refresh token from file or environment variable and returns a new access token
    Returns None if no refresh token is found
    """
    auth_configs = load_auth_config()

    refresh_token = os.environ.get("MINDGARD_API_KEY")
    if not refresh_token and os.path.exists(get_token_file()):
        with open(get_token_file(), "r") as f:
            refresh_token = f.read()

    if refresh_token:
        access_token = (
            requests.post(
                "https://{}/oauth/token".format(auth_configs.AUTH0_DOMAIN),
                data={
                    "grant_type": "refresh_token",
                    "client_id": auth_configs.AUTH0_CLIENT_ID,
                    "audience": auth_configs.AUTH0_AUDIENCE,
                    "refresh_token": refresh_token,
                },
            )
            .json()
            .get("access_token")
        )
        return cast(str, access_token)
    return None


T = TypeVar("T")


# TODO: improve typing definitions here
def require_auth(func: Callable[..., CliResponse]) -> Callable[..., CliResponse]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> CliResponse:
        access_token = load_access_token()
        if not access_token:
            print_to_stderr("\033[1;37mRun `mindgard login`\033[0;0m to authenticate.")
            exit(CliResponse(2).code())
        try:
            return func(*args, **kwargs, access_token=access_token)
        except requests.HTTPError as e:
            if "Unauthorized" in str(e):
                print_to_stderr(
                    "Access token is invalid. Please re-authenticate using `mindgard login`"
                )
                clear_token()
                exit(CliResponse(2).code())
            else:
                raise e

    return wrapper



from typing import Optional
import jsonpath_ng
import jsonpath_ng.ext
import json

from requests import Response

# Filters to only LLM messages with content.
MINIMUM_MESSAGE_SIZE = 10

def extract_reply(json_response,selector:Optional[str]=None, strict=True):
    if selector:
        jsonpath_expr = jsonpath_ng.ext.parse(selector)
        match = jsonpath_expr.find(json_response)
        if match:
            return str(match[0].value)
        elif strict:
            raise Exception(f"Selector {selector} did not match any elements in the response. {json_response=}")
        else:
            return ""
    else:
        return json.dumps(json_response)

class _NDJsonProcessor:
    # Originally based on Ollama streaming responses
    def preprocess_line(self, line:bytes) -> bytes:
        return line

    def join_responses(self, responses: list[str]):
        return "".join(responses)

class _EventStreamProcessor:
    # Originally based on openai streaming responses: https://platform.openai.com/docs/api-reference/streaming
    def preprocess_line(self, line:bytes) -> bytes:
        return line[len('data: '):]

    def join_responses(self, responses: list[str]):
        stripped = [response.strip() for response in responses]
        return " ".join(stripped)

def extract_replies(response:Response, selector:Optional[str]=None):
    content_type = response.headers.get('Content-Type','').lower()
    if (content_type == 'text/event-stream' or content_type == 'application/x-ndjson'):
        reply:list[str] = []
        content_processor = _NDJsonProcessor() if content_type == 'application/x-ndjson' else _EventStreamProcessor()
        for line in response.iter_lines():
            if len(line) < MINIMUM_MESSAGE_SIZE:
                continue
            line_value = content_processor.preprocess_line(line).decode("utf-8")
            line_json = json.loads(line_value)
            extracted = extract_reply(line_json, selector=selector, strict=False)
            if len(extracted) > 0:
                reply.append(extracted)
        return content_processor.join_responses(reply)
    else:
        # Simple RPC
        return extract_reply(response.json(), selector=selector)


from ..run_poll_display import type_ui_exception_map

from typing import Callable, List

from ..wrappers.image import ImageModelWrapper

from rich.progress import Progress

from azure.messaging.webpubsubclient.models import OnGroupDataMessageArgs
from azure.messaging.webpubsubclient import WebPubSubClient

import base64

import logging


def image_message_handler(
    ui_exception_map: type_ui_exception_map,
    ui_exception_progress: Progress,
    submitted_test_id_update: Callable[[str], None],
    model_wrapper: ImageModelWrapper,
    ws_client: WebPubSubClient,
) -> Callable[[OnGroupDataMessageArgs], None]:
    def recv_message_handler(msg: OnGroupDataMessageArgs) -> None:
        if msg.data["messageType"] == "Request":
            logging.debug(f"received request {msg.data=}")
            image_bytes = base64.b64decode(msg.data["payload"]["image"])

            try:
                response = model_wrapper.__call__(image=image_bytes)
            except Exception as e:
                raise e
            finally:
                # we always try to send a response
                replyData = {
                    "correlationId": msg.data["correlationId"],
                    "messageType": "Response",
                    "status": "ok",  # Deprecated but kept for compatibility
                    "payload": {"response": [t.model_dump() for t in response]},
                }
                logging.debug(f"sending response {replyData=}")

                ws_client.send_to_group("orchestrator", replyData, data_type="json")  # type: ignore
        elif msg.data["messageType"] == "StartedTest":
            submitted_test_id_update(msg.data["payload"]["testId"])
        else:
            pass

    return recv_message_handler



from mindgard.wrappers.image import ImageModelWrapper
from mindgard.wrappers.llm import LLMModelWrapper
from ..orchestrator import (
    setup_orchestrator_webpubsub_request,
    OrchestratorSetupRequest,
    OrchestratorTestResponse,
    get_test_by_id, GetTestAttacksResponse,
)
from ..types import (
    type_ui_task_map,
    type_output_func,
    type_submit_func,
    type_ui_exception_map,
)
from ..ui_prefabs import poll_and_display_test, output_test_table

from typing import Optional, List, Callable, Union
from rich.progress import Progress
from rich.table import Table

from azure.messaging.webpubsubclient import WebPubSubClient, WebPubSubClientCredential
from azure.messaging.webpubsubclient.models import OnGroupDataMessageArgs

import time


def model_test_submit_factory(
    request: OrchestratorSetupRequest,
    model_wrapper: Union[LLMModelWrapper, ImageModelWrapper],
    message_handler: Callable[
        [
            type_ui_exception_map,
            Progress,
            Callable[[str], None],
            Union[LLMModelWrapper, ImageModelWrapper],
            WebPubSubClient,
        ],
        Callable[[OnGroupDataMessageArgs], None],
    ],
) -> type_submit_func:

    def model_test_submit(
        access_token: str,
        ui_exception_map: type_ui_exception_map,
        ui_exception_progress: Progress,
    ) -> GetTestAttacksResponse:
        
        response = setup_orchestrator_webpubsub_request(
            access_token=access_token, request=request
        )

        credentials = WebPubSubClientCredential(client_access_url_provider=response.url)
        ws_client = WebPubSubClient(credential=credentials)

        submitted_test_id = [""]

        def update_submited_test_id(id: str) -> None:
            submitted_test_id[0] = id

        recv_message_handler = message_handler(
            ui_exception_map,
            ui_exception_progress,
            update_submited_test_id,
            model_wrapper,
            ws_client,
        )

        ws_client.open()

        ws_client.subscribe("group-message", recv_message_handler)  # type: ignore

        payload = {
            "correlationId": "",
            "messageType": "StartTest",
            "payload": {"groupId": response.group_id},
        }

        ws_client.send_to_group(group_name="orchestrator", content=payload, data_type="json")  # type: ignore

        # wait
        max_attempts = 30
        attempts_remaining = max_attempts
        while submitted_test_id[0] == "":
            time.sleep(1)
            attempts_remaining -= 1
            if attempts_remaining == 0:
                break

        # we're here if submitting was a success...

        if attempts_remaining != 0 and submitted_test_id[0] != "":
            return get_test_by_id(
                test_id=submitted_test_id[0], access_token=access_token
            )
        else:
            raise Exception(
                f"did not receive notification of test submitted within timeout ({max_attempts}s); failed to start test"
            )

    return model_test_submit


def model_test_polling(
    access_token: str,
    initial_test: GetTestAttacksResponse,
    ui_task_map: type_ui_task_map,
    ui_task_progress: Progress,
) -> Optional[GetTestAttacksResponse]:
    return poll_and_display_test(
        access_token,
        ui_task_map,
        ui_task_progress,
        initial_test,
    )


def model_test_output_factory(risk_threshold: int) -> type_output_func:
    def list_llm_test_output(
        test: GetTestAttacksResponse, json_out: bool
    ) -> Optional[Table]:
        return output_test_table(
            json_out=json_out, test=test, risk_threshold=risk_threshold
        )

    return list_llm_test_output



from dataclasses import dataclass
import logging
from typing import Any, Dict, Literal, Optional

import requests

from mindgard.constants import ATTACK_STATE_COMPLETED, ATTACK_STATE_RUNNING, VERSION, ATTACK_STATE_QUEUED

class GeneralException(Exception):
    """
    General exception when interacting with the Mindgard API.
    """

class HttpStatusException(Exception):
    """
    General exception when interacting with the Mindgard API.
    """

@dataclass
class AttackResponse():
    id: str
    name: str
    state: Literal["queued", "running", "completed"]
    errored: Optional[bool] = None
    risk: Optional[int] = None

@dataclass
class FetchTestDataResponse():
    has_finished: bool
    risk: int
    attacks: list[AttackResponse]


def api_response_to_attack_state(attack:Dict[str, Any]) -> AttackResponse:
    if attack["state"] == ATTACK_STATE_QUEUED:
        state = "queued"
    elif attack["state"] == ATTACK_STATE_RUNNING:
        state = "running"
    else:
        state = "completed"

    if attack["state"] < ATTACK_STATE_QUEUED:
        errored = True
    elif attack["state"] == ATTACK_STATE_COMPLETED:
        errored = False
    else:
        errored = None

    risk = attack.get("risk") if attack["state"] == ATTACK_STATE_COMPLETED else None
    return AttackResponse(
        id=attack["id"],
        name=attack["attack"],
        state=state,
        errored=errored,
        risk=risk,
    )

@dataclass
class FetchTestAttacksData():
    has_finished: bool

class MindgardApi():
    def fetch_test_data(
            self, 
            api_base:str, 
            access_token:str, 
            additional_headers:Optional[dict[str, str]],
            test_id:str
    ) -> Optional[FetchTestDataResponse]:
        response = requests.get(
            url=f"{api_base}/assessments/{test_id}",
            headers={
                "Authorization": f"Bearer {access_token}",
                "User-Agent": f"mindgard-cli/{VERSION}",
                "X-User-Agent": f"mindgard-cli/{VERSION}",
                **(additional_headers or {})
            }
        )

        if response.status_code != 200:
            return None
        
        try:
            data = response.json()
        except requests.JSONDecodeError as jde:
            logging.error(f"Error decoding response: {jde}")
            return None

        try:
            attacks = [api_response_to_attack_state(attack) for attack in data["attacks"]]

            return FetchTestDataResponse(
                has_finished=data["hasFinished"],
                attacks=attacks,
                risk=data["risk"],
            )
        except KeyError as ke:
            logging.error(f"KeyError response: {ke}")
            return None
        
    def fetch_test_attacks(
        self, 
        api_base:str, 
        test_id:str,
        access_token:str,
        additional_headers:Optional[dict[str, str]],
    ) -> FetchTestAttacksData:

        try:
            response = requests.get(
                url=f"{api_base}/tests/{test_id}/attacks",
                headers={
                    "Authorization": f"Bearer {access_token}",
                    "User-Agent": f"mindgard-cli/{VERSION}",
                    "X-User-Agent": f"mindgard-cli/{VERSION}",
                    **(additional_headers or {}),
                },
            )
        except requests.RequestException as e:
            raise GeneralException(f"error calling api: {e.__class__.__name__}") from e
        except Exception as e:
            raise GeneralException(f"error calling api: unknown exception") from e

        if response.status_code != 200:
            raise HttpStatusException(f"error calling api. expected 200, got: {response.status_code}")

        try:
            data = response.json()
        except requests.JSONDecodeError as e:
            raise GeneralException(f"error decoding api response: {e}") from e
        
        try:
            return FetchTestAttacksData(
                has_finished=data["test"]["has_finished"],
            )
        except TypeError as e:
            raise GeneralException(f"error parsing api response: {e}") from e


import logging

# OpenAI exceptions

# Typing
from typing import Callable, Dict, Literal, Optional, Type
from requests import status_codes



class MGException(Exception):
    def __init__(self, message: str = "exception") -> None:
        super().__init__(self)
        self.message = message

    def __str__(self) -> str:
        return self.message


# Other Exceptions
class Uncontactable(MGException):
    pass


# HTTP-like Exceptions
class HTTPBaseError(MGException):
    def __init__(self, message: str, status_code: int) -> None:
        super().__init__(message)
        self.status_code: int = status_code
        self.status_message: str = "<unknown>"
        try:
            self.status_message = status_codes._codes[status_code][0]
        except:
            pass


class BadRequest(HTTPBaseError):
    pass


class Unauthorized(HTTPBaseError):
    pass


class Forbidden(HTTPBaseError):
    pass


class NotFound(HTTPBaseError):
    pass


class Timeout(HTTPBaseError):
    pass


class UnprocessableEntity(HTTPBaseError):
    pass


class FailedDependency(HTTPBaseError):
    pass


class RateLimitOrInsufficientCredits(HTTPBaseError):
    pass


class InternalServerError(HTTPBaseError):
    pass


class ServiceUnavailable(HTTPBaseError):
    pass


class EmptyResponse(MGException):
    pass


class NotImplemented(MGException):
    pass


_status_code_exception_map: Dict[int, HTTPBaseError] = {
    400: BadRequest("LLM provider received message that couldn't be handled.", 400),
    401: Unauthorized(
        "User is not authorized to access this LLM provider resource.", 401
    ),
    403: Forbidden(
        "User is known but does not have permission to access this resource.", 403
    ),
    404: NotFound("This resource was not found.", 404),
    408: Timeout("Timed out while trying to access resource", 408),
    422: UnprocessableEntity("Entity sent to LLM could not be processed.", 422),
    424: FailedDependency("Failed Dependency TODO", 424),
    429: RateLimitOrInsufficientCredits("Rate Limit or Insufficient Credits TODO", 429),
    500: InternalServerError("Internal Server Error TODO", 500),
    503: ServiceUnavailable("Service Unavailable TODO", 503),
}


def status_code_to_exception(status_code: int) -> HTTPBaseError:
    return _status_code_exception_map.get(
        status_code, HTTPBaseError("Error specifics unknown", -1)
    )


ErrorCode = Literal[
    "CouldNotContact",
    "ContentPolicy",
    "CLIError",
    "NotImplemented",
    "NoResponse",
    "RateLimited",
    "NetworkIssue",
    "MaxContextLength",
]


exceptions_to_cli_status_codes: Dict[Type[Exception], ErrorCode] = {
    Uncontactable: "CouldNotContact",
    BadRequest: "NoResponse",  # not sure about this, we don't handle 400 atm
    Unauthorized: "NoResponse",
    Forbidden: "NoResponse",
    NotFound: "NoResponse",
    Timeout: "NoResponse",
    UnprocessableEntity: "NoResponse",  # this is currently being handled as a rate limit issue for some reason
    FailedDependency: "NoResponse",
    RateLimitOrInsufficientCredits: "RateLimited",
    InternalServerError: "NoResponse",
    ServiceUnavailable: "NoResponse",
    NotImplemented: "NotImplemented",
    EmptyResponse: "NoResponse",
}


def handle_exception_callback(
    exception: Exception,
    handle_visual_exception_callback: Optional[Callable[[str], None]],
) -> ErrorCode:
    # TODO - come take a look at this
    error_code: ErrorCode = exceptions_to_cli_status_codes.get(type(exception), "CLIError")
    callback_text = str(exception)

    if handle_visual_exception_callback:
        handle_visual_exception_callback(callback_text)

    logging.debug(exception)
    return error_code

